#!/usr/bin/python
""" NAME
        i2c.py

    SYNOPSIS
        python module

    DESCRIPTION
        module to make i2c accesses on the PCIe40 borad lli i2c buses

    OPTIONS
        > ...

    EXAMPLE
        > ...

    AUTHOR
        PY Duval -- 2015

    lICENSE
    (c) Copyright CERN for the benefit of the LHCb Collaboration.
    Distributed under the terms of the GNU General Public Licence V3.
    ana.py script is free and open source software.

    CHANGELOG
        10/02/2017 add the board device number

"""
from ctypes import *
import sys
import getopt

# better to define LD_LIBRARY_PATH to point to the right libs directory
# ===================================================
# use a GLOBAL for library loading and init
# ===================================================
print "loading ", __name__
lli = CDLL("lli-dyn.so")

# DON'T FORGET IT !!!!
# lli.i2c_init(BUS_NB, 125000000, 100000);
# ATTENTION as many values as bus used
NB_OF_BUSES=10
NB_OF_BOARDS=8
initDone = [[False for x in xrange(NB_OF_BUSES)] for x in xrange(NB_OF_BOARDS)]


def init(dev,bus):
    print("init i2c on dev {} bus {}".format(dev, bus))
    lli.i2c_init(dev,bus, 100000000, 100000)
    #lli.i2c_init(bus, 125000000, 100000)
    initDone[dev][bus] = True

#PYD new to set a different clock
# This is the first operation to do if the user doesn't want the default values
def initClock(dev,bus,inCLK, SCLFreq):
    lli.i2c_init(dev,bus, inCLK, SCLFreq)
    initDone[dev][bus] = True


def getInit(dev,bus):
    cval1 = c_int()
    cval2 = c_int()
    status = lli.i2c_getInit(dev,bus, byref(cval1),byref(cval2))
    return (status, cval1.value & 0xffffffff, cval2.value & 0xffffffff,)

def read(dev,bus, add, reg):
    """
    Reads on bus at address add the register number reg
    dev : the board id
    bus : selected i2c bus on the board
    add : address of device on the i2c bus
    reg : number of the register in the device

    returns: list of two values
        status (0 success, -1 error)
        read value

    NOTE: the mask (& 0xffffffff) is necessary to have th full
    unsigned value and not a negative one with high numbers.
    """
    #print "read oni2c bus ", bus, " at address ", hex(add & 0xffffffff) commented by MJ by 19/12/2016
    #print("PYD read dev", dev, " bus ", bus, " add ", hex(add & 0xffffffff), " reg ", reg)
    global initDone
    if initDone[dev][bus] is False:
        init(dev, bus)
    cval = c_int()
    status = lli.i2c_readMem(dev, bus, add, reg, byref(cval))
    if ((cval.value & 0xffffffff) == 0xffffffff):
        print "ERROR: read 0xffffffff at "+hex(add & 0xffffffff)
#    else:
#        print "OK passed read  at 0x%08x value 0x%08x" % \
#              ((reg & 0xffffffff), (cval.value & 0xffffffff))
    return (status, cval.value & 0xffffffff)


def write(dev, bus, add, reg, val):
    """
    Write val on bus at address add in the register number reg
    dev : the board id
    bus : selected i2c bus on the board
    add : address of device on the i2c bus
    reg : number of the register in the device
    val : value to write

    returns: list of two values
        status (0 if success, -1 error)
        written value

    NOTE: the mask (& 0xffffffff) is necessary to have th full
    unsigned value and not a negative one with high numbers
    """

    """
    commented by MJ 19/12/2016
    print "write ", hex(val & 0xffffffff), \
          " on i2c bus ", bus, \
          " at address ", hex(add & 0xffffffff)
    """

    global initDone
    if initDone[dev][bus] is False:
        init(dev, bus)
    cval = c_int(val)
    #print("PYD write dev", dev, " bus ", bus, " add ", hex(add & 0xffffffff), " reg ", reg, " val ", val)
    status = lli.i2c_writeMem(dev,bus, add, reg, byref(cval))
    # LIGNE SUIVANTE RAJOUTTE PAR MJ 040615
#    print "OK passed write  at 0x%08x value 0x%08x" % \
#          ((reg & 0xffffffff), (cval.value & 0xffffffff))
    return (status, cval.value & 0xffffffff)


def main():
    # parse command line options

    dev = 0
    try:
        opts, args = getopt.getopt(sys.argv[1:], "h", ["help"])
    except getopt.error, msg:
        print msg
        print "for help use --help"
        sys.exit(2)
    for o, a in opts:
        if o in ("-h", "--help"):
            print "help"
            sys.exit(0)
    for arg in args:
        print arg  # process() is defined elsewhere

    init(dev,3)
    print "Ini Done"
    print getInit(dev,3)
    print read(dev, 3, 0x0, 0x1)
    print write(dev, 3, 0x0, 0x2, 0x33)


if __name__ == "__main__":
    main()
